import json
import logging
import re
from typing import Optional, List

from semver import VersionInfo

from daktari.check import Check, CheckResult
from daktari.command_utils import get_stdout, can_run_command
from daktari.os import OS
from daktari.version_utils import try_parse_semver
from daktari.checks.google import GkeGcloudAuthPluginInstalled


class KubectlInstalled(Check):
    def __init__(self, required_version: Optional[str] = None, recommended_version: Optional[str] = None):
        self.required_version = required_version
        self.recommended_version = recommended_version
        self.name = "kubectl.installed"
        self.suggestions = {
            OS.OS_X: "<cmd>brew install kubectl</cmd>",
            OS.UBUNTU: "<cmd>sudo snap install kubectl --classic</cmd>",
            OS.GENERIC: "Install kubectl: https://kubernetes.io/docs/tasks/tools/#kubectl",
        }

    def check(self) -> CheckResult:
        installed_version = get_kubectl_version()
        return self.validate_semver_expression(
            "Kubectl", installed_version, self.required_version, self.recommended_version
        )


version_pattern = re.compile("Client Version: v(.*)")


def get_kubectl_version() -> Optional[VersionInfo]:
    raw_version = get_stdout("kubectl version --client=true --short")
    # Handle deprecation of --short
    if raw_version is None:
        raw_version = get_stdout("kubectl version --client=true")

    if raw_version:
        match = version_pattern.search(raw_version)
        if match:
            version = try_parse_semver(match.group(1))
            logging.debug(f"Kubectl version: {version}")
            return version
    return None


class KubectlContextExists(Check):
    depends_on = [GkeGcloudAuthPluginInstalled]

    def __init__(self, context_name: str, provision_command: str = ""):
        self.context_name = context_name
        self.name = f"kubectl.contextExists.{context_name}"
        self.suggestions = {OS.GENERIC: provision_command}

    def check(self) -> CheckResult:
        output = get_stdout("kubectl config get-contexts")
        passed = bool(output and self.context_name in output)
        if not passed:
            return self.failed(f"{self.context_name} is <not/> configured for the current user")

        can_connect = can_run_command(f"kubectl get ns --context {self.context_name}")
        return self.verify(can_connect, f"Could <not/> connect to context {self.context_name}")


class KubectlNoExtraneousContexts(Check):
    def __init__(self, expected_contexts: List[str]):
        self.expected_contexts = expected_contexts
        self.name = "kubectl.noExtraneousContexts"

    def check(self) -> CheckResult:
        output = get_stdout("kubectl config get-contexts -o name")
        if not output:
            return self.failed("Failed to list kubectl contexts")

        contexts = output.splitlines()
        extraneous_contexts = list(filter(lambda context: context not in self.expected_contexts, contexts))
        if extraneous_contexts:
            suggestion = "\n".join(
                f"<cmd>kubectl config delete-context {context}</cmd>" for context in extraneous_contexts
            )
            self.suggestions = {
                OS.GENERIC: suggestion,
            }
            return self.passed_with_warning(f"{len(extraneous_contexts)} extraneous kubectl context(s) found")

        return self.passed("No extraneous kubectl contexts found")


class HelmInstalled(Check):
    name = "helm.installed"

    def __init__(self, required_version: Optional[str] = None, recommended_version: Optional[str] = None):
        self.required_version = required_version
        self.recommended_version = recommended_version
        self.suggestions = {
            OS.OS_X: "<cmd>brew install helm</cmd>",
            OS.UBUNTU: "<cmd>sudo snap install helm --classic</cmd>",
            OS.GENERIC: "Install Helm: https://helm.sh/docs/intro/install/",
        }

    def check(self) -> CheckResult:
        installed_version = get_helm_version()
        return self.validate_semver_expression(
            "Helm", installed_version, self.required_version, self.recommended_version
        )


helm_version_pattern = re.compile("v([0-9\\.]+)")


def get_helm_version() -> Optional[VersionInfo]:
    raw_version = get_stdout("helm version --short")
    if raw_version:
        match = helm_version_pattern.search(raw_version)
        if match:
            version = try_parse_semver(match.group(1))
            logging.debug(f"Helm Version: {version}")
            return version
    return None


class HelmRepoExists(Check):
    depends_on = [HelmInstalled]

    def __init__(self, repo_name: str, repo_url: str):
        self.repo_name = repo_name
        self.repo_url = repo_url.strip("/")
        self.name = f"helm.repoExists.{repo_name}"
        self.suggestions = {
            OS.GENERIC: f"<cmd>helm repo add {repo_name} {repo_url} --force-update</cmd>",
        }

    def check(self) -> CheckResult:
        output = get_stdout("helm repo list -o json")
        if not output:
            return self.failed("No helm repos appear to be configured for the current user.")
        repo_json = json.loads(output)
        repo = next(filter(lambda repo_details: repo_details.get("name") == self.repo_name, repo_json), None)
        if repo is None:
            return self.failed(f"{self.repo_name} is not configured for the current user")

        installed_url = repo["url"].strip("/")
        logging.debug(f"{self.repo_name} helm repo is installed with URL {installed_url}.")

        if installed_url != self.repo_url:
            return self.failed(
                f"{self.repo_name} is configured to use the wrong URL. Expected {self.repo_url}, got {installed_url}"
            )

        return self.passed(f"{self.repo_name} is configured for the current user")



import unittest

from semver import VersionInfo

from daktari.check import CheckResult, CheckStatus
from daktari.test_check_factory import DummyCheck, DummyCheck2

INSTALLED_VERSION = VersionInfo(10, 3, 1)


class TestConfig(unittest.TestCase):
    def test_no_version(self):
        result = DummyCheck().validate_semver_expression("Application", None)
        self._verify_result(result, CheckStatus.FAIL, "Application is not installed")

    def test_invalid_required_version(self):
        result = DummyCheck().validate_semver_expression("Application", INSTALLED_VERSION, required_version=">>zz")
        self._verify_result(result, CheckStatus.FAIL, "Invalid version specification: >zz is not valid SemVer string")

    def test_invalid_recommended_version(self):
        result = DummyCheck().validate_semver_expression("Application", INSTALLED_VERSION, recommended_version=">>zz")
        self._verify_result(result, CheckStatus.FAIL, "Invalid version specification: >zz is not valid SemVer string")

    def test_doesnt_meet_required_version(self):
        result = DummyCheck().validate_semver_expression("Application", INSTALLED_VERSION, required_version=">=10.4.0")
        self._verify_result(result, CheckStatus.FAIL, "Application version is not >=10.4.0 (10.3.1)")

    def test_doesnt_meet_recommended_version(self):
        result = DummyCheck().validate_semver_expression(
            "Application", INSTALLED_VERSION, recommended_version=">=10.4.0"
        )
        self._verify_result(
            result, CheckStatus.PASS_WITH_WARNING, "Application version is 10.3.1, >=10.4.0 recommended"
        )

    def test_meets_all_versions(self):
        result = DummyCheck().validate_semver_expression(
            "Application", INSTALLED_VERSION, required_version=">=10.0.0", recommended_version="<11.0.0"
        )
        self._verify_result(result, CheckStatus.PASS, "Application version is 10.3.1")

    def test_with_dependencies_adds_dependency(self):
        check = DummyCheck(depends_on=None).with_dependencies(DummyCheck, DummyCheck2)
        self.assertEqual(check.depends_on, [DummyCheck, DummyCheck2])

    def test_with_dependencies_keeps_existing_dependencies(self):
        check = DummyCheck(depends_on=[DummyCheck]).with_dependencies(DummyCheck2)
        self.assertEqual(check.depends_on, [DummyCheck, DummyCheck2])

    def _verify_result(self, result: CheckResult, expected_status: CheckStatus, expected_message: str):
        self.assertEqual(result, CheckResult("check.name", expected_status, expected_message, {}))



from importlib_resources import files
from importlib_resources.abc import Traversable


def get_resource(name: str) -> str:
    """Load a textual resource file."""
    return get_resource_path(name).read_text(encoding="utf-8")


def get_resource_path(name: str) -> Traversable:
    return files("daktari.resources").joinpath(name)



from daktari.check import Check, CheckResult
from daktari.os import OS, get_env_var_value


class AndroidNdkHomeSet(Check):
    name = "android.ndkHomeSet"

    def __init__(self, expected_version):
        self.variable_name = "ANDROID_NDK_HOME"
        self.expected_version = expected_version
        self.suggestions = {
            OS.GENERIC: f"""
            Export {self.variable_name} in your shell config.
            The expected value is ANDROID_SDK_HOME/ndk/{self.expected_version}.
            If you manage your android sdk using Android Studio, you can find your ANDROID_SDK_HOME by going to
            Tools > SDK Manager and copying the Android SDK location.
            You may need to check under SDK Tools that you have the NDK installed.
            """
        }

    def check(self) -> CheckResult:
        expected_substring = f"ndk/{self.expected_version}"
        return self.verify(
            expected_substring in get_env_var_value(self.variable_name),
            f"{self.variable_name} is <not/> set with expected version {self.expected_version}",
        )



from typing import Union, Type, Set

from daktari.check import Check
from daktari.collection_utils import flatten


class CyclicCheckException(Exception):
    def __init__(self, message):
        super().__init__(message)


def check_for_cycles(check: Union[Check, Type[Check]], prev_parents: Set[str]):
    if check.name in prev_parents:
        raise CyclicCheckException(f"Check [{check.name}] has cyclic dependencies")
    for sub_check in check.depends_on:
        check_for_cycles(sub_check, prev_parents.union({check.name}))


def get_all_dependent_check_names(check: Union[Check, Type[Check]]) -> Set[str]:
    check_for_cycles(check, set())
    return _get_all_dependent_check_names_recursive(check)


def _get_all_dependent_check_names_recursive(check: Union[Check, Type[Check]]) -> Set[str]:
    sub_dependents = [_get_all_dependent_check_names_recursive(dep) for dep in check.depends_on]
    flat_sub_dependents = flatten(sub_dependents)
    return flat_sub_dependents.union({dep.name for dep in check.depends_on})



import logging
from typing import Optional
from xml.etree.ElementTree import Element, ElementTree, ParseError

from daktari.check import Check, CheckResult
from daktari.file_utils import file_exists


class XmlFileXPathCheck(Check):
    file_path = ""
    xpath_query = "./"
    pass_fail_message = ""

    def validate_query_result(self, result: Optional[Element]) -> bool:
        return result is not None

    def perform_query(self) -> bool:
        if not file_exists(self.file_path):
            logging.debug(f"File {self.file_path} does not exist")
            return False

        try:
            doc = ElementTree(file=self.file_path)
        except ParseError:
            logging.debug(f"Error parsing {self.file_path}", exc_info=True)
            return False

        query_result = doc.find(self.xpath_query)
        return self.validate_query_result(query_result)

    def check(self) -> CheckResult:
        success = self.perform_query()
        return self.verify(success, self.pass_fail_message)



from typing import Optional

from daktari.check import Check, CheckResult
from daktari.command_utils import get_stdout
from daktari.os import OS


class ConanInstalled(Check):
    name = "conan.installed"

    def __init__(self, required_version: Optional[str] = None, recommended_version: Optional[str] = None):
        self.required_version = required_version
        self.recommended_version = recommended_version
        self.suggestions = {OS.GENERIC: "Install conan: <cmd>pip install conan</cmd>"}

    def check(self) -> CheckResult:
        return self.verify_install("conan")


class ConanProfileDetected(Check):
    name = "conan.profileDetected"

    def __init__(self, expected_string: str):
        self.suggestions = {OS.GENERIC: "<cmd>conan profile detect</cmd>"}
        self.expected_string = expected_string
        self.depends_on = [ConanInstalled]

    def check(self) -> CheckResult:
        output = get_stdout("conan profile list")
        expected_profile_detected = output is not None and self.expected_string in output
        return self.verify(expected_profile_detected, f"conan profile {self.expected_string} <not/> detected")



import re
import textwrap
import pyclip
from typing import Callable, Dict, Optional

from colors import green, red, underline, yellow

from daktari.check import CheckResult, CheckStatus
from daktari.os import OS, detect_os


clear_line_prefix = "\33[2K\r"


def check_status_symbol(status: CheckStatus) -> str:
    return {CheckStatus.PASS: "✅", CheckStatus.PASS_WITH_WARNING: "⚠️ ", CheckStatus.FAIL: "❌", CheckStatus.ERROR: "💥"}[
        status
    ]


def check_status_colour(status: CheckStatus) -> Callable:
    return {
        CheckStatus.PASS: green,
        CheckStatus.PASS_WITH_WARNING: yellow,
        CheckStatus.FAIL: red,
        CheckStatus.ERROR: red,
    }[status]


def get_most_specific_suggestion(this_os: str, suggestions: Dict[str, str]) -> Optional[str]:
    return suggestions.get(this_os, suggestions.get(OS.GENERIC))


def print_suggestion_text(text: str):
    text = textwrap.dedent(text.lstrip("\n").rstrip())

    pattern = re.compile("<cmd>(.+)</cmd>")

    def add_ansi_underline(match):
        return underline(match.group(1))

    replaced = pattern.sub(add_ansi_underline, text)
    lines = replaced.splitlines()

    raw_lines = re.compile("</?cmd>").sub("", text).splitlines()

    max_width = max([len(line) for line in raw_lines])

    title = "💡 Suggestion "
    print("┌─" + title + "─" * (max_width - len(title)) + "┐")
    for line in lines:
        print(f"  {line}")
    print("└" + "─" * (max_width + 2) + "┘")


def copy_to_clipboard(suggestion: Optional[str]):
    if suggestion is not None:
        command_regex = re.compile(r"\<cmd\>(.*?)\<\/cmd\>")
        results = command_regex.findall(suggestion)
        if len(results) > 0:
            pyclip.copy("\n".join(results))
            print("ⓘ  Command copied to clipboard")
            return

    print("ⓘ  No command available to copy to clipboard")


def print_check_result(result: CheckResult, early_exit: bool, quiet_mode: bool, idx: int, total_checks: int):
    this_os = detect_os()
    status_symbol = check_status_symbol(result.status)
    colour = check_status_colour(result.status)
    if result.status != CheckStatus.PASS or not quiet_mode:
        print(f"{clear_line_prefix}{status_symbol} [{colour(result.name)}] {result.summary}")
        if result.status in (CheckStatus.FAIL, CheckStatus.PASS_WITH_WARNING):
            suggestion = get_most_specific_suggestion(this_os, result.suggestions)
            if suggestion:
                print_suggestion_text(suggestion)

            if early_exit:
                copy_to_clipboard(suggestion)

        if quiet_mode:
            print("")

    if quiet_mode:
        print_progress_bar(early_exit, idx + 1, total_checks)
    elif early_exit:
        print("ⓘ  Exited early due to --fail-fast flag")


def print_progress_bar(early_exit: bool, current: int, total: int):
    end_char = "\n" if current == total or early_exit else "\r"
    print(progress_bar(current, total, early_exit), end=end_char)


def progress_bar(current: int, total: int, early_exit: bool) -> str:
    fraction = current / total

    arrow_head = "x" if early_exit else ">"
    arrow = int(fraction * 25 - 1) * "-" + arrow_head
    padding = int(25 - len(arrow)) * " "

    return f"Progress: [{arrow}{padding}] {int(fraction*100)}%  ({current}/{total})"



import logging
import re
from typing import Optional

from semver import VersionInfo

from daktari.check import Check, CheckResult
from daktari.command_utils import get_stdout
from daktari.os import OS
from daktari.version_utils import try_parse_semver


class DockerInstalled(Check):
    name = "docker.installed"

    suggestions = {
        OS.GENERIC: "Install docker: https://docs.docker.com/get-docker/",
        OS.OS_X: "Install docker: https://docs.docker.com/docker-for-mac/install/",
    }

    def __init__(self, required_version: Optional[str] = None):
        self.required_version = required_version

    def check(self) -> CheckResult:
        installed_version = get_docker_version()
        return self.validate_semver_expression("Docker", installed_version, self.required_version)


major_version_pattern = re.compile("Docker version ([0-9.]+)")


def get_docker_version() -> Optional[VersionInfo]:
    raw_version = get_stdout("docker --version")
    if raw_version:
        match = major_version_pattern.search(raw_version)
        if match:
            version_string = match.group(1)
            version = try_parse_semver(version_string)
            logging.debug(f"Docker version - raw: {version_string}, parsed: {version}")
            return version
    return None


class DockerComposeInstalled(Check):
    name = "docker-compose.installed"

    suggestions = {OS.GENERIC: "Install docker-compose: https://docs.docker.com/compose/install/"}

    def check(self) -> CheckResult:
        return self.verify_install("docker-compose")



from daktari.check import Check, CheckResult
from daktari.os import OS


class PythonInstalled(Check):
    def __init__(self, required_version: int):
        self.required_version = required_version
        self.name = f"python{required_version}.installed"
        self.suggestions = {
            OS.OS_X: f"<cmd>brew install python{required_version}</cmd>",
            OS.UBUNTU: f"<cmd>sudo apt install python{required_version}-dev</cmd>",
            OS.GENERIC: f"Download python {required_version}: https://www.python.org/downloads/",
        }

    def check(self) -> CheckResult:
        return self.verify_install(f"python{self.required_version}")



import unittest

from daktari.check import CheckStatus
from daktari.checks.misc import HostAliasesConfigured


class TestHostAliasesConfigured(unittest.TestCase):
    def test_checking_hosts_does_not_blow_up_on_success(self):
        result = HostAliasesConfigured({}).check()
        self.assertEqual(result.status, CheckStatus.PASS)

    def test_checking_hosts_does_not_blow_up_on_failure(self):
        result = HostAliasesConfigured({"host": "no.such.entry.surely"}).check()
        self.assertEqual(result.status, CheckStatus.FAIL)


if __name__ == "__main__":
    unittest.main()
